"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family category data processor class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.Data.Objects.ifamily_processor import IFamilyProcessor
from duHast.Revit.Categories.Data.Objects import category_data as rCatData
from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Utilities.Objects import result as res
from duHast.Revit.Categories.Utility import category_property_names as rCatPropNames


class CategoryProcessor(IFamilyProcessor):
    def __init__(self, pre_actions=None, post_actions=None):
        """
        Class constructor.
        """

        # setup report header
        key_prefix = (
            rCatData.GRAPHIC_PROPERTY_KEY_PREFIX
            + rCatData.GRAPHIC_PROPERTY_KEY_PREFIX_DELIMITER
        )
        string_report_headers = [
            IFamData.ROOT,
            IFamData.ROOT_CATEGORY,
            IFamData.FAMILY_NAME,
            IFamData.FAMILY_FILE_PATH,
            IFamData.USAGE_COUNTER,
            IFamData.USED_BY,
            rCatData.CATEGORY_NAME,
            rCatData.SUB_CATEGORY_NAME,
            rCatData.SUB_CATEGORY_ID,
            key_prefix + rCatPropNames.CATEGORY_GRAPHIC_STYLE_3D,
            key_prefix + rCatPropNames.CATEGORY_GRAPHIC_STYLE_CUT,
            key_prefix + rCatPropNames.CATEGORY_GRAPHIC_STYLE_PROJECTION,
            key_prefix + rCatPropNames.PROPERTY_MATERIAL_NAME,
            key_prefix + rCatPropNames.PROPERTY_MATERIAL_ID,
            key_prefix + rCatPropNames.PROPERTY_LINE_WEIGHT_CUT_NAME,
            key_prefix + rCatPropNames.PROPERTY_LINE_WEIGHT_PROJECTION_NAME,
            key_prefix + rCatPropNames.PROPERTY_LINE_COLOUR_RED_NAME,
            key_prefix + rCatPropNames.PROPERTY_LINE_COLOUR_GREEN_NAME,
            key_prefix + rCatPropNames.PROPERTY_LINE_COLOUR_BLUE_NAME,
        ]

        # store data type  in base class
        super(CategoryProcessor, self).__init__(
            pre_actions=pre_actions,
            post_actions=[self._post_action_update_used_subcategories],
            data_type="Category",
            string_report_headers=string_report_headers,
        )

        # self.data = []
        # self.dataType = 'Category'

        # list of corner cases when it comes to category checking: imports in families or Reference planes ... ( english language specific!! )
        self.category_check_corner_cases = ["Imports in Families", "Reference Planes"]

        # self.preActions = preActions
        # set default post action to updated categories used in root processor with any categories found in nested
        # families
        # self.postActions = [self._postActionUpdateUsedSubcategories]
        # add any other post actions
        if post_actions != None:
            for p_action in post_actions:
                self.post_actions.append(p_action)

    def process(self, doc, root_path, root_category_path):
        """
        Calls processor instance with the document and root path provided and adds processor instance to class property .data

        :param doc: Current family document.
        :type doc: Autodesk.Revit.DB.Document
        :param root_path: The path of the nested family in a tree: rootFamilyName::nestedFamilyNameOne::nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type root_path: str
        :param root_category_path: The category path of the nested family in a tree: rootFamilyCategory::nestedFamilyOneCategory::nestedFamilyTwoCategory\
            This includes the actual family category as the last node.
        :type root_category_path: str
        """

        dummy = rCatData.CategoryData(root_path, root_category_path, self.data_type)
        dummy.process(doc)
        self.data.append(dummy)

    # --------------------------------------------- post action ----------------------------------------------------------

    def _add_data(
        self,
        processor,
        root,
        root_category_path,
        fam_name,
        fam_path,
        use_counter,
        used_by,
        fam_cat_name,
        sub_cat_name,
        sub_cat_id,
        cat_gra_style_three_d,
        cat_gra_style_cut,
        cat_gra_style_pro,
        prop_mat_name,
        prop_mat_id,
        prop_line_weight_cut_name,
        prop_line_weight_projection_name,
        prop_line_col_red,
        prop_line_col_green,
        prop_line_col_blue,
    ):

        processor.add_Data(
            root,
            root_category_path,
            fam_name,
            fam_path,
            use_counter,
            used_by,
            fam_cat_name,
            sub_cat_name,
            sub_cat_id,
            cat_gra_style_three_d,
            cat_gra_style_cut,
            cat_gra_style_pro,
            prop_mat_name,
            prop_mat_id,
            prop_line_weight_cut_name,
            prop_line_weight_projection_name,
            prop_line_col_red,
            prop_line_col_green,
            prop_line_col_blue,
        )

    def _is_sub_category_present(self, root_family_data, nested_family_sub_category):
        match = None
        # check whether sub category is present
        for root_fam in root_family_data:
            if (
                root_fam[rCatData.CATEGORY_NAME]
                == nested_family_sub_category[rCatData.CATEGORY_NAME]
                and root_fam[rCatData.SUB_CATEGORY_NAME]
                == nested_family_sub_category[rCatData.SUB_CATEGORY_NAME]
            ):
                match = root_fam
                break
        return match

    def _update_root_family_data(
        self, root_family_data, nested_families_sub_categories
    ):
        # loop over nested family subcategory data
        for nested_sub_category in nested_families_sub_categories:
            # check if sub category is already in root family
            matching_root_fam_category = self._is_sub_category_present(
                root_family_data, nested_sub_category
            )
            if matching_root_fam_category != None:
                # update used by list
                if (
                    nested_sub_category[IFamData.FAMILY_NAME]
                    not in matching_root_fam_category[IFamData.USED_BY]
                ):
                    # add the root path to the used by list for ease of identification of the origin of this subcategory usage
                    matching_root_fam_category[IFamData.USED_BY].append(
                        nested_sub_category[IFamData.ROOT]
                    )
                    # update used by counter
                    matching_root_fam_category[IFamData.USAGE_COUNTER] = (
                        matching_root_fam_category[IFamData.USAGE_COUNTER] + 1
                    )
            else:
                pass
                # nothing to do if that category has not been reported to start off with
                # this category could, for example, belong to the section marker family present in most 3d families

    def _get_used_subcategories(self, data):
        used_subcategories = []
        for d in data:
            if d[IFamData.USAGE_COUNTER] > 0:
                # get the family category
                category_path = d[IFamData.ROOT_CATEGORY].split(" :: ")
                # which is the last entry in the root category path
                category = category_path[len(category_path) - 1]
                # select only items which either belong to the category of the family or
                # are corner cases like imports in families or Reference planes ... ( english language specific!! )
                if (
                    category == d[rCatData.CATEGORY_NAME]
                    or d[rCatData.CATEGORY_NAME] in self.category_check_corner_cases
                ):
                    used_subcategories.append(d)
        return used_subcategories

    def _post_action_update_used_subcategories(self, doc):
        return_value = res.Result()
        try:
            # find all subcategories of nested families
            nested_family_data = self._find_nested_families_data()
            # get used sub categories from nested data
            nested_family_used_sub_categories = self._get_used_subcategories(
                nested_family_data
            )
            # update root family data only
            root_family_data = self._find_root_family_data()
            # update root processor data as required
            self._update_root_family_data(
                root_family_data, nested_family_used_sub_categories
            )
            return_value.update_sep(
                True, "Post Action Update subcategories data successful completed."
            )
        except Exception as e:
            return_value.update_sep(
                False,
                "Post Action Update subcategories data failed with exception: {}".format(
                    e
                ),
            )
        return return_value



import glob
from os import path


def get_link_path(fileName, possibleLinkLocations, fileExtension):
    """
    Gets a fully qualified file path to a file name match (revit project file extension .rvt) in given directory locations.
    Returns the first file name match it finds! If no match found returns None.
    :param fileName: Filter to identify match. Filter is string.startswith(fileName)
    :type fileName: str
    :param possibleLinkLocations: List of folders which may contain the link file.
    :type possibleLinkLocations: list str
    :param fileExtension: A file extension in format: '.xyz'. Use '.rvt' for revit project files.
    :type fileExtension: str
    :return: Fully qualified file path if match is found otherwise None.
    :rtype: str
    """

    linkPath = None
    counter = 0
    try:
        foundMatch = False
        # attempt to find filename match in given locations
        for linkLocation in possibleLinkLocations:
            fileList = glob.glob(linkLocation + "\\*" + fileExtension)
            if fileList != None:
                for file in fileList:
                    fileNameInFolder = path.basename(file)
                    if fileNameInFolder.startswith(fileName):
                        linkPath = file
                        counter = +1
                        foundMatch = True
                        break
        # return none if multiple matches where found
        if foundMatch == True and counter > 1:
            linkPath = None
    except Exception:
        linkPath = None
    return linkPath



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains string padding functions for message formatting. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Sample output without a time stamp:

------------------------------header------------------------------

or with a time stamp:

23-04-30 20_31_07 : -----------------header------------------------ 
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities import date_stamps

#: how long is a padded row
DEFAULT_PADDING_LENGTH = 90


def pad_header_no_time_stamp(header_name, padding_length=DEFAULT_PADDING_LENGTH):
    """
    Pads a header string to be centred in a row of dashes of a given length.

    :param header_name: The header
    :type header_name: str
    :param padding_length: the length of the padded header excluding time stamp, defaults to 70
    :type padding_length: int, optional
    :return: A padded header row.
    :rtype: str
    """
    if padding_length > len(header_name) + 2:
        sides = (padding_length - len(header_name)) // 2
        # return the padded header text
        return "\n" + "-".ljust(sides, "-") + header_name + "-".ljust(sides, "-")
    else:
        return "\n" + header_name


def pad_header_with_time_stamp(header_name, padding_length=DEFAULT_PADDING_LENGTH):
    """
    Pads a header string to be centred in a row of dashes of a given length.

    Will also add a time stamp to the beginning of the message.

    :param header_name: The header
    :type header_name: str
    :param padding_length: the length of the padded header excluding time stamp, defaults to 70
    :type padding_length: int, optional
    :return: A padded header row.
    :rtype: str
    """

    # get a time stamp
    time_stamp = date_stamps.get_date_stamp("%y-%m-%d %H_%M_%S : ")
    # check whether max length is not exceeded
    if padding_length > len(header_name) + 2 + len(time_stamp):
        sides = (padding_length - len(header_name) - len(time_stamp)) // 2
        # return the padded header text
        return (
            "\n"
            + date_stamps.get_date_stamp("%y-%m-%d %H_%M_%S : ")
            + "-".ljust(sides, "-")
            + header_name
            + "-".ljust(sides, "-")
        )
    else:
        return "\n" + header_name


def pad_string(message, padding_length=DEFAULT_PADDING_LENGTH):
    """
    Pads a string message to be formatted: left hand side message, right hand side status (if any)
    Maximum length 70 characters (excludes time stamp!)
    If message is longer then 70-2 characters it will be returned un-changed.

    :param message: The message to be padded
    :type message: str
    :param padding_length: Length of message string after padding, defaults to 70
    :type padding_length: int, optional
    :return: Padded message
    :rtype: str
    """

    if len(message) < padding_length:
        status_length = 0
        status = ""
        if "[False]" in message:
            status_length = len("[False]")
            status = "[False]"
        elif "[True]" in message:
            status_length = len("[True]")
            status = "[True]"
        if status_length > 0:
            message_left = message[:-status_length]
            message_left = message_left.ljust(padding_length - status_length, ".")
            message = message_left + status
            return message
        else:
            return message
    else:
        return message



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit family change category helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

from duHast.Revit.Categories.Utility.category_properties_get_utils import get_category_graphic_style_ids, get_category_properties
from duHast.Revit.Categories.Utility.elements_by_category_utils import get_elements_by_category, move_elements_to_category
from duHast.Revit.Categories.categories import get_main_sub_categories, set_family_category
from duHast.Revit.Categories.family_sub_categories import create_new_category_from_saved_properties
from duHast.Utilities.Objects import result as res


def change_family_category(doc, new_category_name):
    """
    Changes the current family category to the new one specified.

    Revit's default behavior when changing the category of a family is to discard all custom subcategories created and assign elements which are on those custom subcategories to the new family category. This function will also re-create any user-created subcategories under the new category and assign elements to them to match the subcategory they were on before the category change.

    :param doc: Current Revit family document.
    :type doc: Autodesk.Revit.DB.Document
    :param new_category_name: The new family category.
    :type new_category_name: str

    :return: Result class instance.
        - `result.status` (bool): True if all custom subcategories were re-created under the new family category and elements were moved to those subcategories, otherwise False.
        - `result.message` (str): Confirmation of successful creation of subcategories and element move.
        - `result.result` (list): Empty list.

    On exception:
        - `result.status` (bool): False.
        - `result.message` (str): Generic exception message.
        - `result.result` (list): Empty.

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # get sub categories in family
    sub_cats = get_main_sub_categories(doc)

    # get all elements on custom subcategories
    elements = {}
    for sub_cat in sub_cats:
        el = get_elements_by_category(doc, sub_cats[sub_cat])
        elements[sub_cat] = el

    # get properties of all custom sub categories
    props = {}
    for sub_cat in sub_cats:
        prop = get_category_properties(sub_cats[sub_cat], doc)
        props[sub_cat] = prop

    # change family category
    change_fam = set_family_category(doc, new_category_name)
    return_value.update(change_fam)

    if change_fam.status:
        # re-create custom sub categories
        for sub_cat in sub_cats:
            # only re-create custom sub categories (id greater then 0)
            if sub_cats[sub_cat].Id.IntegerValue > 0:
                # create new sub categories with flag: ignore if cut graphic style is missing set to true!
                create_cat = create_new_category_from_saved_properties(
                    doc, sub_cat, props[sub_cat], True
                )
                return_value.update(create_cat)
                if create_cat.status:
                    # get the graphic style ids of the new subcategory for elements to use
                    destination_cat_ids = get_category_graphic_style_ids(
                        create_cat.result
                    )
                    # move elements back onto custom subcategories
                    move_el = move_elements_to_category(
                        doc, elements[sub_cat], sub_cat, destination_cat_ids
                    )
                    return_value.update(move_el)
                else:
                    return_value.update(create_cat)
    else:
        return_value.update_sep(
            False, "Failed to change family category: {}".format(change_fam.message)
        )
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family Base data analysis module containing functions to build a reload tree.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- read change list:
-   text file (tab separated) with columns:familyName	familyFilePath	categoryName
- read overall family base data list
- get direct parents of change list families
- get next level parents of direct parents

- loop direct parent list until empty:
-   - remove any direct parents which also exist in next level parents
-   - write direct parents to file
-   - set next level parents as direct parents
-   - find all direct parents of changed families in base data list

- those reload lists will then be separated into work chunks by file list writer...

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Utilities.Objects.timer import Timer
from duHast.Utilities.Objects import result as res

from duHast.Revit.Family.Data import family_base_data_utils as rFamBaseDataUtils

from duHast.Revit.Family.Data.family_reload_advanced_utils import (
    read_change_list,
    write_out_empty_task_list,
    write_reload_list_to_file,
)


def _remove_duplicates(list_one, list_two):
    """
    Removes any item from list one which is present in list two.

    :param list_one: List of tuples containing root family data.
    :type list_one: [rFamBaseDataUtils.rootFamily]
    :param list_two: List of tuples containing root family data.
    :type list_two: [rFamBaseDataUtils.rootFamily]

    :return: _description_
    :rtype: _type_
    """

    new_list = []
    duplicates_list = []
    for l_one_item in list_one:
        if l_one_item not in list_two:
            new_list.append(l_one_item)
        else:
            duplicates_list.append(l_one_item)
    return new_list, duplicates_list


def _get_hosts(
    current_families, overall_family_base_nested_data, overall_family_base_root_data
):
    """
    Returns the direct ( one level up) host families of the current families.

    :param current_families: A list of current families represented as tuples (tuple need to have properties 'name' and 'category').
    :type current_families: [rFamBaseDataUtils.rootFamily] or [rFamBaseDataUtils.nestedFamily] or [rFamReloadAdvUtils.changedFamily]
    :param overall_family_base_nested_data: List of tuples containing nested family data.
    :type overall_family_base_nested_data: [rFamBaseDataUtils.nestedFamily]
    :param overall_family_base_root_data: List of tuples containing root family data.
    :type overall_family_base_root_data: [rFamBaseDataUtils.rootFamily]

    :return: A list of root families.
    :rtype: [rFamBaseDataUtils.rootFamily]
    """

    # get current change list host files
    direct_hosts = rFamBaseDataUtils.find_root_families_from_hosts(
        rFamBaseDataUtils.find_all_direct_host_families(
            current_families, overall_family_base_nested_data
        ),
        overall_family_base_root_data,
    )
    return direct_hosts


def build_work_lists(
    change_list_file_path,
    family_base_data_report_file_path,
    load_lists_output_directory_path,
):
    """
    Processes a file change list and a family base data report. From both reports it builds a lists for reloading families bottom up in their nesting hierarchy.

    :param change_list_file_path: Fully qualified file path to family change list report file.
    :type change_list_file_path: str
    :param family_base_data_report_file_path: Fully qualified file path to family base data report file.
    :type family_base_data_report_file_path: str
    :param load_lists_output_directory_path: Fully qualified directory path to which the task output files will be written
    :type load_lists_output_directory_path: str
    :raises Exception: "Infinite loop." Will be raised if more then 20 task output files are written (representing a family nesting level of 20 deep...unlikely)
    """

    # set up a timer
    t_process = Timer()
    t_process.start()

    return_value = res.Result()
    change_list = read_change_list(change_list_file_path)
    return_value.append_message(
        "{} Change list of length [{}] loaded.".format(
            t_process.stop(), len(change_list)
        )
    )

    t_process.start()
    # read overall family base data from file
    (
        overall_family_base_root_data,
        overall_family_base_nested_data,
    ) = rFamBaseDataUtils.read_overall_family_data_list(
        family_base_data_report_file_path
    )
    return_value.append_message(
        "{} Nested base data list of length [{}] loaded.".format(
            t_process.stop(), len(overall_family_base_nested_data)
        )
    )

    if len(change_list) > 0:
        # list containing the hosts of the host families
        task_next_level = []
        # safety switch in case of infinite loop
        task_list_counter = 0

        t_process.start()
        task_current_level = _get_hosts(
            change_list, overall_family_base_nested_data, overall_family_base_root_data
        )
        return_value.append_message(
            "{} Direct hosts [{}] found.".format(
                t_process.stop(), len(task_current_level)
            )
        )

        t_process.start()
        task_next_level = _get_hosts(
            task_current_level,
            overall_family_base_nested_data,
            overall_family_base_root_data,
        )
        return_value.append_message(
            "{} Next level hosts [{}] found.".format(
                t_process.stop(), len(task_next_level)
            )
        )

        # loop until no more entries in current level tasks
        while len(task_current_level) > 0:

            # remove next level hosts from direct hosts list to avoid overlap in reload process
            cleaned_current_tasks, over_lap_tasks = _remove_duplicates(
                task_current_level, task_next_level
            )

            # write out cleaned up list:
            if len(cleaned_current_tasks) > 0:
                t_process.start()
                result_write_to_disk = write_reload_list_to_file(
                    cleaned_current_tasks,
                    load_lists_output_directory_path,
                    task_list_counter,
                )
                return_value.update_sep(
                    result_write_to_disk,
                    "{} Wrote task list to file with status: {}".format(
                        t_process.stop(), result_write_to_disk
                    ),
                )
            else:
                # write out an empty task list!
                empty_task_list_flag = write_out_empty_task_list(
                    load_lists_output_directory_path, task_list_counter
                )
                return_value.update_sep(
                    empty_task_list_flag,
                    "Wrote empty task list at counter [{}]".format(task_list_counter),
                )

            # swap lists to get to next level of loading
            task_current_level = list(task_next_level)
            return_value.append_message(
                "Swapping next level hosts to direct hosts [{}]".format(
                    len(task_current_level)
                )
            )

            t_process.start()
            # get next level host families (task)
            task_next_level = _get_hosts(
                task_current_level,
                overall_family_base_nested_data,
                overall_family_base_root_data,
            )
            return_value.append_message(
                "{} Next level hosts [{}]".format(
                    t_process.stop(), len(task_next_level)
                )
            )

            # increase task list counter to be used in file name
            task_list_counter = task_list_counter + 1
            if task_list_counter > 20:
                # trigger fail save
                return_value.update_sep(
                    False, " Exceeded maximum number of task list files! (20)"
                )
                raise Exception("Infinite loop.")
    else:
        return_value.update_sep(
            True, "Empty change list found. No families require processing."
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit warnings. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

# -------------------------------------------- common variables --------------------


# --------------------------------------------- utility functions ------------------


def get_warnings(doc):
    """
    Returns a list of warnings from the model

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of all failure messages in model.
    :rtype: list of Autodesk.Revit.DB.FailureMessage
    """

    return doc.GetWarnings()


def get_warnings_by_guid(doc, guid):
    """
    Returns all failure message objects where failure definition has matching GUID

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param guid: Filter: Identifying a specific failure of which the corresponding messages are to be returned.
    :type guid: Autodesk.Revit.DB.Guid

    :return: list of all failure messages with matching guid
    :rtype: list of Autodesk.Revit.DB.FailureMessage
    """

    filtered_warnings = []
    warnings = doc.GetWarnings()
    for warning in warnings:
        if str(warning.GetFailureDefinitionId().Guid) == guid:
            filtered_warnings.append(warning)
    return filtered_warnings


def get_warnings_grouped_by_relation(doc, guid):
    """
    Returns a dictionary of warnings where all warnings specified by guid related to each other are grouped together.
    Key will be the element with the lowest element id, value will be a list of element ids, in ascending order, of all warnings in the group (including the key!)

    Note:
    Revit must report exactly two elements per warning for this to work

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param guid: Filter: Identifying a specific failure of which the corresponding messages are to be returned.
    :type guid: Autodesk.Revit.DB.Guid

    :return: Dictionary of warnings grouped by element id
    :rtype: dict
    """

    # get all warning relating to a guid
    warnings = get_warnings_by_guid(doc, guid)

    warning_grouping = {}
    counter = 0
    for warning in warnings:
        match = False
        element_ids = warning.GetFailingElements()

        # just in case there are more than 2 elements in the warning
        if len(element_ids) != 2:
            continue

        # initialise the first warning group
        if len(warning_grouping) == 0:
            warning_grouping[counter] = [
                element_ids[0].IntegerValue,
                element_ids[1].IntegerValue,
            ]
            continue

        # check if the warning relates to an existing group
        for key, value in warning_grouping.items():
            if (
                element_ids[0].IntegerValue in value
                and element_ids[1].IntegerValue not in value
            ):
                warning_grouping[key].append(element_ids[1].IntegerValue)
                # print("added {} to group {}".format( element_ids[1].IntegerValue, counter))
                match = True
                break
            elif (
                element_ids[1].IntegerValue in value
                and element_ids[0].IntegerValue not in value
            ):
                warning_grouping[key].append(element_ids[0].IntegerValue)
                # print("added {} to group {}".format( element_ids[0].IntegerValue, counter))
                match = True
                break
            elif (
                element_ids[1].IntegerValue in value
                and element_ids[0].IntegerValue in value
            ):
                # print("matched {} {} to group {} {}".format( element_ids[0].IntegerValue, element_ids[1].IntegerValue, counter, value))
                match = True
                break
            else:
                pass
                # print("no match for {} {} in {}".format(element_ids[0].IntegerValue, element_ids[1].IntegerValue, value))
        if not match:
            counter += 1
            warning_grouping[counter] = [
                element_ids[0].IntegerValue,
                element_ids[1].IntegerValue,
            ]
            # print("new counter group: {} to group {}, {}".format(counter, element_ids[0].IntegerValue, element_ids[1].IntegerValue,))
            # new warning grouping is required

    sorted_warning_grouping = {}
    # sort the values in the dictionary
    for key, value in warning_grouping.items():
        warning_grouping[key] = sorted(value)
        sorted_warning_grouping[sorted(value)[0]] = sorted(value)
    return sorted_warning_grouping



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to comma separated text files. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import codecs
import csv

from duHast.Utilities.files_io import get_file_name_without_ext


def get_unique_headers(files):
    """
    Gets a list of alphabetically sorted headers retrieved from text files.
    Assumes:

    - first row in each file is the header row
    - headers are separated by <tab> character

    :param files: List of file path from which the headers are to be returned.
    :type files: list of str
    :return: List of headers.
    :rtype: list of str
    """

    headers_in_all_files = {}
    for f in files:
        # get unmodified row data and remove the next line character at the end
        data = get_first_row_in_csv_file(f)
        headers_in_all_files[get_file_name_without_ext(f)] = data
    headers_unique = []
    for header_by_file in headers_in_all_files:
        empty_header_counter = 0
        for header in headers_in_all_files[header_by_file]:
            # reformat any empty headers to be unique
            if header == "":
                header = header_by_file + ".Empty." + str(empty_header_counter)
                empty_header_counter = empty_header_counter + 1
            if header not in headers_unique:
                headers_unique.append(header)
    return sorted(headers_unique)


def read_csv_file(filepathCSV, increaseMaxFieldSizeLimit=False):
    """
    Read a csv file into a list of rows, where each row is another list.
    :param filepathCSV: The fully qualified file path to the csv file.
    :type filepathCSV: str
    :return: A list of list of strings representing the data in each row.
    :rtype: list of list of str
    """

    row_list = []

    # hard coded hack
    if increaseMaxFieldSizeLimit:
        csv.field_size_limit(2147483647)

    try:
        with open(filepathCSV) as csv_file:
            reader = csv.reader(csv_file)
            for row in reader:  # each row is a list
                row_list.append(row)
            csv_file.close()
    except Exception as e:
        print(str(e))
        row_list = []
    return row_list


def get_first_row_in_csv_file(filePath):
    """
    Reads the first line of a csv text file and returns it as a list of strings
    :param filePath: The fully qualified file path.
    :type filePath: str
    :return: The first row of a text file.
    :rtype: str
    """

    return_value = []
    try:
        with open(filePath) as csv_file:
            reader = csv.reader(csv_file)
            for row in reader:  # each row is a list
                return_value = row
                break
            csv_file.close()
    except Exception as e:
        print(str(e))
    return return_value


def write_report_data_as_csv(file_name, header, data, write_type="w"):
    """
    Function writing out report information as CSV file.
    :param file_name: The reports fully qualified file path.
    :type file_name: str
    :param header: list of column headers
    :type header: list of str
    :param data: list of list of strings representing row data
    :type data: [[str,str,..]]
    :param write_type: Flag indicating whether existing report file is to be overwritten 'w' or appended to 'a', defaults to 'w'
    :type write_type: str, optional
    """

    # open the file in the write mode
    with codecs.open(file_name, write_type, encoding="utf-8") as f:
        # create the csv writer
        writer = csv.writer(f)
        # check header
        if len(header) > 0:
            writer.writerow(header)
        if len(data) > 0:
            for d in data:
                # write a row to the csv file
                writer.writerow(d)
        f.close()



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions to rename family loaded families in a project file or family file.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This helper function expect a folder containing rename directive files. For format of those files refer to module RevitFamilyRenameFilesUtils


"""


#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

# import clr
# import System

from duHast.Revit.Family import family_rename_files_utils as rFamRenameUtils
from duHast.Revit.Family import family_utils as rFamUtils
from duHast.Revit.Common import transaction as rTran
from duHast.Utilities.Objects import result as res

# import Autodesk Revit DataBase namespace
from Autodesk.Revit.DB import Transaction


def _rename_loaded_families(doc, rename_directives, family_ids):
    """
    Loops over nested families and if a match in rename directives is found will rename the family accordingly.

    :param doc: The current family document.
    :type doc: Autodesk.Revit.DB.Document
    :param rename_directives: List of rename directives.
    :type rename_directives: [rename_directive]
    :param family_ids: List of all nested family ids.
    :type family_ids: [Autodesk.Revit.DB.ElementId]

    :return:
        Result class instance.

        - result.status. True if all families where renamed successfully, otherwise False.
        - result.message will contain each rename messages in format 'Renamed family from :' +current Name + ' to ' + newName.
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message in format: 'Failed to rename family from :' + currentName + ' to ' + newName
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    return_value.status = False
    rename_match_counter = 0
    # loop over families and check for match in rename directives
    for fam_id in family_ids:
        family = doc.GetElement(fam_id)
        family_name = family.Name
        if family.IsEditable and family.IsValidObject:
            family_category_name = family.FamilyCategory.Name
            # loop over rename directives and look for match in family name and category
            for rename_directive in rename_directives:
                if (
                    rename_directive.name == family_name
                    and rename_directive.category == family_category_name
                ):
                    rename_match_counter = rename_match_counter + 1
                    # rename this family
                    def action():
                        action_return_value = res.Result()
                        try:
                            family.Name = rename_directive.newName
                            action_return_value.update_sep(
                                True,
                                "Renamed family of category ["
                                + family_category_name
                                + "] vs directive category ["
                                + rename_directive.category
                                + "] from: "
                                + rename_directive.name
                                + " to: "
                                + rename_directive.newName,
                            )
                        except Exception as e:
                            action_return_value.update_sep(
                                False,
                                "Failed to rename family of category ["
                                + family_category_name
                                + "] vs directive category ["
                                + rename_directive.category
                                + "] from: "
                                + rename_directive.name
                                + " to: "
                                + rename_directive.newName,
                            )
                        return action_return_value

                    transaction = Transaction(
                        doc, "Renaming: {}".format(rename_directive.name)
                    )
                    rename_result = rTran.in_transaction(transaction, action)
                    if rename_result.status:
                        # make sure that this returns true as soon as one family renamed successfully
                        return_value.status = True
                    # update messages
                    return_value.append_message(rename_result.message)
                    break
    # check if anything got renamed at all
    if rename_match_counter == 0:
        return_value.append_message(
            "No match for rename directives found. Nothing was renamed."
        )
    return return_value


def rename_loaded_families(doc, directory_path):
    """
    Entry point for this module. Will read rename directives files in given directory and attempt to rename
    loaded families accordingly.

    :param directory_path: Fully qualified directory path to where rename directive files are located.
    :type directory_path: str
    :return:
        Result class instance.

        - result.status. True if a single families was renamed successfully, otherwise False.
        - result.message will contain each rename messages in format 'Renamed family from :' +current Name + ' to ' + newName.
        - result.result empty list

        On exception:

        - result.status (bool) will be False.
        - result.message will contain an exception message in format: 'Failed to rename family from :' + currentName + ' to ' + newName
        - result.result will be empty

    :rtype: :class:`.Result`
    """

    return_value = res.Result()
    # get directives from folder
    rename_directives_result = rFamRenameUtils.get_rename_directives(directory_path)
    # check if anything came back
    if rename_directives_result.status:
        rename_directives = rename_directives_result.result
        # get all family ids in file
        family_ids = rFamUtils.get_all_loadable_family_ids_through_types(doc)
        if len(family_ids) > 0:
            # rename files as per directives
            return_value = _rename_loaded_families(doc, rename_directives, family_ids)
        else:
            return_value.update_sep(True, "Mo loadable families in file.")
    else:
        return_value = rename_directives_result

    return return_value



"""
A class used to store Revit warnings information.

This class represents Revit warnings and provides attributes to store information about them, such as the filename, warning ID, description, and associated element IDs.
"""


#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects import base
from duHast.Utilities.date_stamps import (
    get_date_stamp,
    FILE_DATE_STAMP_YYYYMMDD_SPACE,
    TIME_STAMP_HHMMSEC_COLON,
)

class RevitWarning(base.Base):
    """
    A class representing a warning in a Revit file. This class inherits from the base.Base class.

    Attributes:
        - file_name (str): The name of the Revit file associated with the warning.
        - id (str): The ID of the warning.
        - description (str): The description of the warning.
        - element_ids (list of str): A list of element IDs associated with the warning.
        
    This class is used to store information about warnings that occur in a Revit file, such as the filename, warning ID, description, and associated element IDs.
    """


    def __init__(self, file_name="", id="", description="", element_ids=[], **kwargs):
        """
        Initializes a RevitWarning object.

        Args:
            file_name (str, optional): The name of the Revit file associated with the warning. Defaults to "".
            id (str, optional): The ID of the warning. Defaults to "".
            description (str, optional): The description of the warning. Defaults to "".
            element_ids (list, optional): A list of element IDs associated with the warning. Defaults to [].
            `**kwargs` (optional): Additional keyword arguments to be passed to the base.Base class constructor.
        """


        super(RevitWarning, self).__init__(**kwargs)

        self.file_name = file_name
        self.id = id
        self.description = description
        self.element_ids = element_ids
        self.date = get_date_stamp(FILE_DATE_STAMP_YYYYMMDD_SPACE),
        self.time = get_date_stamp(TIME_STAMP_HHMMSEC_COLON)
    
    
    def class_to_csv(self, headers):
        """
        Converts the warning object to a CSV list based on the provided headers.

        Args:
            headers (list): A list of headers for the CSV list.

        Returns:
            list: The warning object converted to a CSV list.
        """

        if isinstance(self, object):
            csv_list = []
            for prop in headers:
                if(prop in self.__dict__):
                    csv_list.append(self.__dict__[prop])
                else:
                    csv_list.append("Property {} does not exist!".format(prop))
            return csv_list
        else:
            return []



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit families helper functions to change the reference type of reference planes and curve based elements.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import Autodesk.Revit.DB as rdb

from duHast.Revit.Common import (
    parameter_get_utils as rParaGet,
    parameter_set_utils as rParaSet,
)
from duHast.Utilities.Objects import result as res
from duHast.Revit.Family.family_element_utils import (
    get_all_curve_based_elements_in_family,
)


def set_ref_planes_to_not_a_reference(doc):
    """
    This function sets any reference plane with reference type 'weak' within a family to reference type 'not a reference'.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: Result class instance.
        - `result.status` (bool): True if at least one reference plane type was successfully changed; otherwise, False.
        - `result.message` (str): One row entry per reference plane requiring reference type change.
        - `result.result` (not used).

    :rtype: :class:`.Result`
    """

    """
    Revit API reference types and their integer values:
    - ('ref name', 'Left', 'reference type as int', 0, 'reference type as string', 'Left')
    - ('ref name', 'Center (Left/Right)', 'reference type as int', 1, 'reference type as string', 'Center (Left/Right)')
    - ('ref name', 'Right', 'reference type as int', 2, 'reference type as string', 'Right')
    - ('ref name', 'Front', 'reference type as int', 3, 'reference type as string', 'Front')
    - ('ref name', 'Reference Plane', 'reference type as int', 4, 'reference type as string', 'Center (Front/Back)')
    - ('ref name', 'Back', 'reference type as int', 5, 'reference type as string', 'Back')
    - ('ref name', 'Reference Plane', 'reference type as int', 6, 'reference type as string', 'Bottom')
    - ('ref name', 'Reference Plane', 'reference type as int', 7, 'reference type as string', 'Center (Elevation)')
    - ('ref name', 'Top', 'reference type as int', 8, 'reference type as string', 'Top')
    - ('ref name', 'Reference Plane', 'reference type as int', 12, 'reference type as string', 'Not a Reference')
    - ('ref name', 'Reference Plane', 'reference type as int', 13, 'reference type as string', 'Strong Reference')
    - ('ref name', 'Reference Plane', 'reference type as int', 14, 'reference type as string', 'Weak Reference')
    """


    result = res.Result()
    result.update_sep(True, "Changing reference status of reference planes...")
    match_at_all = False
    collector_ref_planes = rdb.FilteredElementCollector(doc).OfClass(rdb.ReferencePlane)
    for ref_p in collector_ref_planes:
        value_int = rParaGet.get_built_in_parameter_value(
            ref_p,
            rdb.BuiltInParameter.ELEM_REFERENCE_NAME,
            rParaGet.get_parameter_value_as_integer,
        )
        # check if an update is required (id is greater then 12)
        if value_int > 13:
            result_change = rParaSet.set_built_in_parameter_value(
                doc, ref_p, rdb.BuiltInParameter.ELEM_REFERENCE_NAME, "12"
            )
            # set overall flag to indicate that at least one element was changed
            if result_change.status == True and match_at_all == False:
                match_at_all = True
            result.update(result_change)
    if match_at_all == False:
        result.update_sep(False, "No reference planes found requiring reference type update")
    return result


def set_symbolic_and_model_lines_to_not_a_reference(doc):
    """
    This function sets any model or symbolic curve in a family with reference type 'weak' to reference type 'not a reference'.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: Result class instance.
        - `result.status` (bool): True if at least one curve reference type was successfully changed; otherwise, False.
        - `result.message` (str): One row entry per curve element requiring reference type change.
        - `result.result` (not used).

    :rtype: :class:`.Result`
    """

    """
    Revit API reference types and their integer values:
    - ('ref name', 'Model Lines', 'reference type as int', 0, 'reference type as string', 'Not a Reference')
    - ('ref name', 'Model Lines', 'reference type as int', 1, 'reference type as string', 'Weak Reference')
    - ('ref name', 'Model Lines', 'reference type as int', 2, 'reference type as string', 'Strong Reference')
    - ('ref name', 'Symbolic Lines', 'reference type as int', 0, 'reference type as string', 'Not a Reference')
    - ('ref name', 'Symbolic Lines', 'reference type as int', 1, 'reference type as string', 'Weak Reference')
    - ('ref name', 'Symbolic Lines', 'reference type as int', 2, 'reference type as string', 'Strong Reference')
    """


    result = res.Result()
    result.update_sep(True, "Changing reference status of model and symbolic curves...")
    match_at_all = False
    curves = get_all_curve_based_elements_in_family(doc)
    for curve in curves:
        # get the current reference type
        value_int = rParaGet.get_built_in_parameter_value(
            curve,
            rdb.BuiltInParameter.ELEM_IS_REFERENCE,
            rParaGet.get_parameter_value_as_integer,
        )
        # check if an update is required (id equals 1)
        if value_int == 1:
            result_change = rParaSet.set_built_in_parameter_value(
                doc, curve, rdb.BuiltInParameter.ELEM_IS_REFERENCE, "0"
            )
            # set overall flag to indicate that at least one element was changed
            if result_change.status == True and match_at_all == False:
                match_at_all = True
            result.update(result_change)
    if match_at_all == False:
        result.status = False
        result.message = "No curve elements found requiring reference type update"
    return result



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit family instance placement helper functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


# class used for stats reporting
from duHast.Utilities.Objects import result as res
from duHast.Revit.Common.transaction import in_transaction

from Autodesk.Revit.DB import (
    Transaction,
)


from Autodesk.Revit.DB.Structure import StructuralType


def place_a_family_instance_by_level(
    doc,
    location_point,
    family_symbol,
    target_placement_level,
    transaction_manager=in_transaction,
    modify_action=None,
):
    """
    Places a family instance by level.

    Note: It is much faster to modify the family instance after placement in the same transaction than to place it in one transaction
    and modify its parameters in a separate transaction.
    Use the modify_action parameter to provide a function that modifies the family instance after placement. Function to accept a family instance as a parameter.

    :param doc: The revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param location_point: The location point of the family instance.
    :type location_point: Autodesk.Revit.DB.XYZ
    :param family_symbol: The family symbol to place.
    :type family_symbol: Autodesk.Revit.DB.FamilySymbol
    :param target_placement_level: The level to place the family instance on.
    :type target_placement_level: Autodesk.Revit.DB.Level
    :param transaction_manager: The transaction manager to use.
    :type transaction_manager: function
    :param modify_action: An action to run after the family instance is placed to modify its properties.
    :type modify_action: function

    :return: A result object.
    :rtype: duHast.Utilities.Objects.result.Result
    """

    return_value = res.Result()

    # places a family in the model
    def action():
        action_return_value = res.Result()
        try:
            # the symbol needs to be activated before using it
            # https://thebuildingcoder.typepad.com/blog/2014/08/activate-your-family-symbol-before-using-it.html
            if not family_symbol.IsActive:
                family_symbol.Activate()
                doc.Regenerate()
                
            # create a new instance
            element = doc.Create.NewFamilyInstance(
                location_point,
                family_symbol,
                target_placement_level,
                StructuralType.NonStructural,
            )

            action_return_value.append_message("Family placed successfully.")
            action_return_value.result.append(element)

            # if a modify action is provided, run it
            if modify_action:
                modify_result = modify_action(element)
                if isinstance(modify_result, res.Result):
                    action_return_value.update(modify_result)
            else:
                action_return_value.append_message(
                    "No modify action provided, skipping."
                )

        except Exception as e:
            action_return_value.update_sep(
                False, "Failed to place family instance: {}".format(e)
            )
        return action_return_value

    transaction = Transaction(doc, "Placing Family")
    return_value = transaction_manager(transaction, action)

    return return_value

def place_a_family_instance_in_basic_wall(doc,
    location_point,
    wall,
    family_symbol,
    target_placement_level,
    transaction_manager=in_transaction,
    modify_action=None,
    ):
    """
    Places a family instance in a basic wall.

    Note: It is much faster to modify the family instance after placement in the same transaction than to place it in one transaction
    and modify its parameters in a separate transaction.
    Use the modify_action parameter to provide a function that modifies the family instance after placement. Function to accept a family instance as a parameter.

    :param doc: The revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param location_point: The location point of the family instance.
    :type location_point: Autodesk.Revit.DB.XYZ
    :param wall: The wall to place the family instance in.
    :type wall: Autodesk.Revit.DB.Wall
    :param family_symbol: The family symbol to place.
    :type family_symbol: Autodesk.Revit.DB.FamilySymbol
    :param target_placement_level: The level to place the family instance on.
    :type target_placement_level: Autodesk.Revit.DB.Level
    :param transaction_manager: The transaction manager to use.
    :type transaction_manager: function
    :param modify_action: An action to run after the family instance is placed to modify its properties.
    :type modify_action: function

    :return: A result object.
    :rtype: duHast.Utilities.Objects.result.Result
    """

    # places a family in the model
    def action():
        action_return_value = res.Result()
        try:
            # the symbol needs to be activated before using it
            # https://thebuildingcoder.typepad.com/blog/2014/08/activate-your-family-symbol-before-using-it.html
            if not family_symbol.IsActive:
                family_symbol.Activate()
                doc.Regenerate()
            # create a new instance
            element = doc.Create.NewFamilyInstance(
                location_point,
                family_symbol,
                wall,
                target_placement_level,
                StructuralType.NonStructural,
            )
            
            action_return_value.append_message("Family placed successfully.")
            action_return_value.result.append(element)

            # if a modify action is provided, run it
            if modify_action:
                modify_result = modify_action(element)
                if isinstance(modify_result, res.Result):
                    action_return_value.update(modify_result)
            else:
                action_return_value.append_message(
                    "No modify action provided, skipping."
                )

        except Exception as e:
            action_return_value.update_sep(
                False, "Failed to place family instance: {}".format(e)
            )
        return action_return_value

    transaction = Transaction(doc, "Placing Family")
    return_value = transaction_manager(transaction, action)
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit view templates. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.Common import common as com
from duHast.Revit.Views.Utility.view_types import _get_view_types


def get_view_templates(doc):
    """
    Get all view templates in a model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: All view templates in the model
    :rtype: list of Autodesk.Revit.DB.View
    """

    view_templates = []
    col = rdb.FilteredElementCollector(doc).OfClass(rdb.View)
    for v in col:
        # filter out templates
        if v.IsTemplate:
            view_templates.append(v)
    return view_templates


def get_view_templates_ids(doc):
    """
    Get all view template ids in a model
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: All view templates Id's in the model
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col = rdb.FilteredElementCollector(doc).OfClass(rdb.View)
    for v in col:
        # filter out templates
        if v.IsTemplate:
            ids.append(v.Id)
    return ids


def get_used_view_templates_ids(doc):
    """
    Gets ids of view templates used in views in the model only
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: All used view templates Id's in the model
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    view_template_ids_used = []
    # get all view templates assigned to views
    col = rdb.FilteredElementCollector(doc).OfClass(rdb.View)
    for v in col:
        # filter out browser organization and other views which cant be deleted
        if (
            v.IsTemplate == False
            and v.ViewType != rdb.ViewType.SystemBrowser
            and v.ViewType != rdb.ViewType.ProjectBrowser
            and v.ViewType != rdb.ViewType.Undefined
            and v.ViewType != rdb.ViewType.Internal
            and v.ViewType != rdb.ViewType.DrawingSheet
        ):
            if (
                v.ViewTemplateId not in view_template_ids_used
                and v.ViewTemplateId != rdb.ElementId.InvalidElementId
            ):
                view_template_ids_used.append(v.ViewTemplateId)
    return view_template_ids_used


def get_default_view_type_template_ids(doc):
    """
    Gets view template Id's used as default by view types
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: All view templates Id's which are used as default in view types in the model
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    view_template_ids_used = []
    # get all templates assigned to view family types:
    view_family_templates = com.get_similar_type_families_by_type(doc, _get_view_types)
    for vt in view_family_templates:
        for id in vt[1]:
            # get the element
            vt_fam = doc.GetElement(id)
            if (
                vt_fam.DefaultTemplateId not in view_template_ids_used
                and vt_fam.DefaultTemplateId != rdb.ElementId.InvalidElementId
            ):
                view_template_ids_used.append(vt_fam.DefaultTemplateId)
    return view_template_ids_used


def get_all_used_view_template_ids(doc):
    """
    Get all used view template Id's.

    Templates can either be:
    - used as default by view types
    - used by a view

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: All view templates Id's which are used in the model.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    vtv = get_used_view_templates_ids(doc)
    view_family_templates = get_default_view_type_template_ids(doc)
    for id in view_family_templates:
        if id not in vtv:
            vtv.append(id)
    return vtv


def get_template_ids_which_can_have_filters(doc, filter_by_type):
    """
    Get all templates in a model of given type

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param filter_by_type: List of view types of which to return view templates from
    :type filter_by_type: list of Autodesk.Revit.DB.ViewType
    :return: All view templates in the model
    :rtype: list of Autodesk.Revit.DB.View
    """

    view_templates = []
    col = rdb.FilteredElementCollector(doc).OfClass(rdb.View)
    for v in col:
        # filter out templates
        if v.IsTemplate:
            for filter in filter_by_type:
                if v.ViewType == filter:
                    view_templates.append(v)
                    break
    return view_templates


def get_all_unused_view_template_ids(doc):
    """
    Gets all view template Id's not used by view types or by views
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: All view templates Id's which are not used in the model.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """
    used_vts = get_all_used_view_template_ids(doc)
    vt_in_model = get_view_templates(doc)
    unused_vts = []
    for vt in vt_in_model:
        if vt.Id not in used_vts:
            unused_vts.append(vt.Id)
    return unused_vts



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Common Revit API utility functions.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#!/usr/bin/python
# -*- coding: utf-8 -*-
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

# import datetime
import System
import clr


clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

# import glob
# class used for stats reporting

# import everything from Autodesk Revit DataBase namespace (Revit API)
from Autodesk.Revit.DB import (
    BuiltInParameter,
    BuiltInCategory,
    FilteredElementCollector,
)
import os.path as path

# utilities
from duHast.Utilities import utility as util
from duHast.Revit.Common import parameter_get_utils as rParaGet

# importing revit groups module
from duHast.Revit.Common import groups as rGroup


def get_element_mark(e):
    """
    Returns the mark value of an element.

    :param e: The element.
    :type e: Autodesk.Revit.DB.Element

    :return:
        The element mark value.
        If an exception occurred, the message will be 'Failed with exception: ' + the exception string.
    :rtype: str
    """

    mark = ""
    try:
        para_mark = e.get_Parameter(BuiltInParameter.ALL_MODEL_MARK)
        mark = "" if para_mark == None else para_mark.AsString()
    except Exception as e:
        mark = "Failed with exception: " + str(e)
    return mark


def element_ids_to_semicolon_delimited_text(element_ids):
    """
    Convert a list of ElementIds to a semicolon delimited string

    :param element_ids: List of ElementIds
    :type element_ids: List[ElementId]
    :return: Semicolon delimited string of ElementIds
    :rtype: str
    """
    return str.Join("; ", [str(elementId.IntegerValue) for elementId in element_ids])


# ----------------------------------------Legend Components -----------------------------------------------


def get_legend_components_in_model(doc, type_ids):
    """
    Returns all symbol (type) ids of families which have been placed as legend components and have match in list past in.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param type_ids: List of typeIds to check against.
    :type type_ids: list str
    :raise: Any exception will need to be managed by the function caller.

    :return: Values are representing symbol (type) ids of legend components in models filtered by ids past in.
    :rtype: list of str
    """

    ids = []
    # get all legend components in the model to check against list past in
    col = FilteredElementCollector(doc).OfCategory(BuiltInCategory.OST_LegendComponents)
    for c in col:
        id = rParaGet.get_built_in_parameter_value(
            c, BuiltInParameter.LEGEND_COMPONENT, rParaGet.get_parameter_value
        )
        if id in type_ids and id not in ids:
            ids.append(id)
            break
    return ids


# ----------------------------------------types - Autodesk.Revit.DB ElementType -----------------------------------------------


def get_similar_type_families_by_type(doc, type_getter):
    """
    Returns a list of unique types its similar family (symbol) types.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param function type_getter:
        The function which takes the document as an argument and returns a list of family symbols (types).
    :raise: Any exception will need to be managed by the function caller.

    :return: list of Autodesk.Revit.DB.Symbol and Autodesk.Revit.DB.ElementId:
    :rtype: List [[Autodesk.Revit.DB.ElementType, Autodesk.Revit.DB.ElementId, Autodesk.Revit.DB.ElementId,...],]
    """

    sim_types = []
    types = type_getter(doc)
    for t in types:
        t_data = [t]
        sims = t.GetSimilarTypes()
        sim_data = []
        for sim in sims:
            sim_data.append(sim)
        # simData.sort() # not sure a sort is actually doing anything
        t_data.append(sim_data)
        if check_unique_type_data(sim_types, t_data):
            sim_types.append(t_data)
    return sim_types


def check_unique_type_data(existing_types, new_type_data):
    """
    Compares two lists of types and their similar types (ids).

    Assumes that second list past in has only one occurrence of type and its similar types
    Compares types by name and if match their similar types.

    :param existing_types: Source list
    :type existing_types: List of List in format [[Autodesk.Revit.DB.ElementType , Autodesk.Revit.DB.ElementId, Autodesk.Revit.DB.ElementId,...],]
    :param new_type_data: Comparison list
    :type new_type_data: List in format [Autodesk.Revit.DB.ElementType, Autodesk.Revit.DB.ElementId, Autodesk.Revit.DB.ElementId,...]

    :return:
        True, if new type is not in list existing Types passed in or
        if ids of similar family types do not match any similar types already in list
    :rtype: bool
    """

    result = True
    for s in existing_types:
        # check for matching family name
        if s[0].FamilyName == new_type_data[0].FamilyName:
            # check if match has the same amount of similar family types
            # if not it is unique
            if len(s[1]) == len(new_type_data[1]):
                # assume IDs do match
                match_i_ds = True
                for i in range(len(s[1])):
                    if s[1][i] != new_type_data[1][i]:
                        # id's dont match, this is unique
                        match_i_ds = False
                        break
                if match_i_ds:
                    # data is not unique
                    result = False
                    break
    return result


def get_unused_type_ids_in_model(doc, type_getter, instance_getter):
    """
    Returns ID of unused family types in the model.

    Used in purge code since it leaves at least one type behind (built in families require at least one type in the model)

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param type_getter: Function accepting current document as argument and returning a collector of types in model
    :type type_getter: func
    :param instance_getter: Function accepting current document as argument and returning a list of instances in model
    :type instance_getter: func

    :return: List of type ids which can be purged from the model.
    :rtype: list Autodesk.Revit.DB.ElementId
    """

    # get all  types available and associated family types
    family_types_available = get_similar_type_families_by_type(doc, type_getter)
    # get used type ids
    used_family_type_ids = instance_getter(doc)
    # flag indicating that at least one type was removed from list because it is in use
    # this flag used when checking how many items are left...
    removed_at_least_one = []
    # set index to 0, type names might not be unique!!
    counter = 0
    # loop over available types and check which one is used
    for t in family_types_available:
        # remove all used family type Id's from the available list...
        # whatever is left can be deleted if not last available item in list for type
        # there should always be just one match
        for used_family_type_id in used_family_type_ids:
            # get the index of match
            index = util.index_of(t[1], used_family_type_id)
            # remove used item from list
            if index > -1:
                t[1].pop(index)
                if t not in removed_at_least_one:
                    removed_at_least_one.append(counter)
        counter = counter + 1
    # filter these by family types where is only one left
    # make sure to leave at least one family type behind, since the last type cannot be deleted
    filtered_unused_type_ids = []
    # reset index
    counter = 0
    for t in family_types_available:
        if counter in removed_at_least_one:
            # at least one item was already removed from list...so all left over ones can be purged
            for id in t[1]:
                # get the element
                t_fam = doc.GetElement(id)
                if t_fam.CanBeDeleted:
                    filtered_unused_type_ids.append(id)
        else:
            # need to keep at least one item
            if len(t[1]) > 1:
                # maxLength = len(t[1])
                # make sure to leave the first one behind to match Revit purge behavior
                for x in range(1, len(t[1])):
                    id = t[1][x]
                    # get the element
                    t_fam = doc.GetElement(id)
                    # check whether this can be deleted...
                    if t_fam.CanBeDeleted:
                        filtered_unused_type_ids.append(id)
        counter = counter + 1
    return filtered_unused_type_ids


# ----------------------------------------instances of types - Autodesk.Revit.DB ElementType -----------------------------------------------


def get_not_placed_types(doc, get_types, get_instances):
    """

    returns a list of unused types foo by comparing type Ids of placed instances with types past in.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param get_types: Types getter function. Needs to accept doc as argument and return a collector of type foo
    :type get_types: func (doc)
    :param get_instances: Instances getter function. Needs to accept doc as argument and return a collector of instances foo
    :type get_instances: func(doc)

    :return: returns a list of unused types
    :rtype: list of type foo
    """

    avail_types = get_types(doc)
    placed_instances = get_instances(doc)
    not_placed = []
    already_checked = []
    # loop over all types and check for matching instances
    for at in avail_types:
        match = False
        for pi in placed_instances:
            # check if we had this type checked already, if so ignore and move to next
            if pi.GetTypeId() not in already_checked:
                #  check for type id match
                if pi.GetTypeId() == at.Id:
                    # add to already checked and verified as match list
                    already_checked.append(pi.GetTypeId())
                    match = True
                    break
        if match == False:
            not_placed.append(at)
    return not_placed


# --------------------------------------------- check whether groups contain certain element types - Autodesk.Revit.DB ElementType  ------------------


def check_group_for_type_ids(doc, group_type, type_ids):
    """

    Filters passed in list of type ids by type ids found in group and returns list of unmatched Id's

    This only returns valid data if at least one instance of the group is placed in the model, otherwise GetMemberIds() returns empty!!

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param group_type: Group to be checked whether they contains elements of types past in.
    :type group_type: Autodesk.Revit.DB.GroupType
    :param type_ids: List of type ids to confirm whether they are in use a group
    :type type_ids: list of Autodesk.Revit.Db.ElementId

    :return: Returns all type ids not matched
    :rtype: list of Autodesk.Revit.Db.ElementId
    """

    unused_type_ids = []
    used_type_ids = []
    # get the first group from the group type and get its members
    for g in group_type.Groups:
        # get ids of group elements:
        member_ids = g.GetMemberIds()
        # built list of used type ids
        for member_id in member_ids:
            member = doc.GetElement(member_id)
            used_type_id = member.GetTypeId()
            if used_type_id not in used_type_ids:
                used_type_ids.append(used_type_id)
    for check_id in type_ids:
        if check_id not in used_type_ids:
            unused_type_ids.append(check_id)
    return unused_type_ids


def check_groups_for_matching_type_ids(doc, group_types, type_ids):
    """
    Checks all elements in groups past in whether group includes element of which type Id is matching any type ids past in

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param group_types: Groups to be checked whether they contains elements of types past in.
    :type group_types: list of Autodesk.Revit.DB.GroupType
    :param type_ids: List of type ids to confirm whether they are in use a group
    :type type_ids: list of Autodesk.Revit.Db.ElementId

    :return: Returns all type ids not matched
    :rtype: list of Autodesk.Revit.Db.ElementId
    """

    for group_type in group_types:
        type_ids = check_group_for_type_ids(doc, group_type, type_ids)
        # check if all type ids where matched up
        if len(type_ids) == 0:
            break
    return type_ids


def get_unused_type_ids_from_detail_groups(doc, type_ids):
    """
    Checks elements in nested detail groups and detail groups whether their type ElementId is in the list past in.

    This only returns valid data if at least one instance of the group is placed in the model!!!

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param type_ids: List of type ids to confirm whether they are in use a group
    :type type_ids: list of Autodesk.Revit.Db.ElementId

    :return: Returns all type Ids from list past in not found in group definitions
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    unused_type_ids = []
    nested_detail_groups = rGroup.get_nested_detail_groups(doc)
    detail_groups = rGroup.get_detail_groups(doc)
    unused_type_ids = check_groups_for_matching_type_ids(
        doc, nested_detail_groups, type_ids
    )
    unused_type_ids = check_groups_for_matching_type_ids(doc, detail_groups, type_ids)
    return unused_type_ids


# ----------------------------------------elements-----------------------------------------------


def get_ids_from_element_collector(col):
    """
    This will return a list of all element ids in collector.

    Any element in collector which is invalid will be ignored.

    :param col: A filtered element collector.
    :type col: Autodesk.Revit.DB.FilteredElementCollector

    :return: list of all element ids of valid elements in collector.
    :rtype: List of Autodesk.Revit.DB.ElementId
    """

    ids = []
    for c in col:
        try:
            ids.append(c.Id)
        except Exception as e:
            pass
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family category data class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Utilities import utility as util
from duHast.Revit.Categories import categories as rCat
from duHast.Revit.Categories.Utility import category_properties_get_utils as rCatPropGet
from duHast.Revit.Categories.Utility import category_property_names as rCatPropNames
from duHast.Revit.Categories.Utility import (
    elements_by_category_utils as rElementByCatUtils,
)


# import Autodesk
# import Autodesk.Revit.DB as rdb

# data dictionary key values specific to this class
CATEGORY_NAME = "categoryName"
SUB_CATEGORY_NAME = "subCategoryName"
SUB_CATEGORY_ID = "subCategoryId"
GRAPHIC_PROPERTY_KEY_PREFIX = "graphicProperty"
GRAPHIC_PROPERTY_KEY_PREFIX_DELIMITER = "_"


class CategoryData(IFamData.IFamilyData):
    def __init__(self, root_path=None, root_category_path=None, data_type=None):
        """
        Class constructor

        :param root_path: The path of the nested family in a tree: rootFamilyName::nestedFamilyNameOne::nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type root_path: str
        :param data_type: Human readable data type descriptor
        :type data_type: str
        """

        # store data type  in base class
        super(CategoryData, self).__init__(
            root_path=root_path,
            root_category_path=root_category_path,
            data_type=data_type,
        )

    def add_data(
        self,
        root,
        root_category_path,
        fam_name,
        fam_path,
        use_counter,
        used_by,
        fam_cat_name,
        sub_cat_name,
        sub_cat_id,
        cat_gra_style_three_d,
        cat_gra_style_cut,
        cat_gra_style_pro,
        prop_mat_name,
        prop_mat_id,
        prop_line_weight_cut_name,
        prop_line_weight_projection_name,
        prop_line_col_red,
        prop_line_col_green,
        prop_line_col_blue,
    ):

        dic = {
            IFamData.ROOT: root,
            IFamData.ROOT_CATEGORY: root_category_path,
            IFamData.FAMILY_NAME: fam_name,
            IFamData.FAMILY_FILE_PATH: fam_path,
            IFamData.USAGE_COUNTER: use_counter,
            IFamData.USED_BY: used_by,
            CATEGORY_NAME: fam_cat_name,
            SUB_CATEGORY_NAME: sub_cat_name,
            SUB_CATEGORY_ID: sub_cat_id,
            rCatPropNames.CATEGORY_GRAPHIC_STYLE_3D: cat_gra_style_three_d,
            rCatPropNames.CATEGORY_GRAPHIC_STYLE_CUT: cat_gra_style_cut,
            rCatPropNames.CATEGORY_GRAPHIC_STYLE_PROJECTION: cat_gra_style_pro,
            rCatPropNames.PROPERTY_MATERIAL_NAME: prop_mat_name,
            rCatPropNames.PROPERTY_MATERIAL_ID: prop_mat_id,
            rCatPropNames.PROPERTY_LINE_WEIGHT_CUT_NAME: prop_line_weight_cut_name,
            rCatPropNames.PROPERTY_LINE_WEIGHT_PROJECTION_NAME: prop_line_weight_projection_name,
            rCatPropNames.PROPERTY_LINE_COLOUR_RED_NAME: prop_line_col_red,
            rCatPropNames.PROPERTY_LINE_COLOUR_GREEN_NAME: prop_line_col_green,
            rCatPropNames.PROPERTY_LINE_COLOUR_BLUE_NAME: prop_line_col_blue,
        }

        self.data.append(dic)

    def _create_data(
        self,
        root,
        root_category_path,
        fam_name,
        fam_path,
        use_counter,
        used_by,
        fam_cat_name,
        sub_cat_name,
        sub_cat_id,
        category_graphic_properties,
    ):
        """
        Generates dictionary object from data past in.

        CategoryGraphicProperties will be flattened with prefix 'graphicProperty'

        :param root: The path of the nested family in a tree: rootFamilyName::nestedFamilyNameOne::nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type root: str
        :param root_category_path: The path of the nested family in a tree of categories. This includes the actual family category as the last node.
        :type root_category_path: str
        :param fam_name: The family name
        :type fam_name: str
        :param fam_path: The family saved file path
        :type fam_path: str
        :param use_counter: Counter of how many objects in the family are of this category
        :type use_counter: int
        :param used_by: List of element ids of elements of this category.
        :type used_by: [Autodesk.Revit.DB.ElementId]
        :param fam_cat_name: The families category.
        :type fam_cat_name: str
        :param sub_cat_name: The subcategory name.
        :type sub_cat_name: str
        :param sub_cat_id: The subcategory id.
        :type sub_cat_id: Autodesk.Revit.DB.ElementId
        :param category_graphic_properties: List of dictionaries describing the graphic properties of this category.
        :type category_graphic_properties: [dict]
        
        :return: A flatt dictionary describing key values and properties of a category.
        :rtype: dict
        """

        dic = {
            IFamData.ROOT: root,
            IFamData.ROOT_CATEGORY: root_category_path,
            IFamData.FAMILY_NAME: fam_name,
            IFamData.FAMILY_FILE_PATH: fam_path,
            IFamData.USAGE_COUNTER: use_counter,
            IFamData.USED_BY: used_by,
            CATEGORY_NAME: fam_cat_name,
            SUB_CATEGORY_NAME: sub_cat_name,
            SUB_CATEGORY_ID: sub_cat_id,
        }

        # flatten dictionary
        for d in category_graphic_properties:
            dummy = util.flatten(
                d, GRAPHIC_PROPERTY_KEY_PREFIX, GRAPHIC_PROPERTY_KEY_PREFIX_DELIMITER
            )
            dic.update(dummy)

        return dic

    def _build_data(self, main_sub_cats, main_cat_name, doc):
        """
        Extracts for each category past in, its properties and usage and adds that data to class property .data

        :param main_sub_cats: List of sub categories to be processed.
        :type main_sub_cats: [Autodesk.Revit.DB.Category]
        :param main_cat_name: The parent category name.
        :type main_cat_name: str
        :param doc: Current family document
        :type doc: Autodesk.Revit.DB.Document
        """

        # get usage of each main sub category
        for key, subCat in main_sub_cats.items():
            # get elements using category
            element_dic = rElementByCatUtils.get_elements_by_category(doc, subCat)
            # get category property
            cat_props = rCatPropGet.get_category_properties(subCat, doc)
            # add element counter for 3D, Cut, Elevation style
            use_counter = 0
            for key in element_dic:
                use_counter = use_counter + len(element_dic[key])
            # add element ids integer value in 3D, Cut, Elevation style
            used_by_ids = []
            for key in element_dic:
                for id in element_dic[key]:
                    used_by_ids.append(id.IntegerValue)

            # build data dictionary
            dic = self._create_data(
                self.root_path,
                self.root_category_path,
                self._strip_file_extension(doc.Title),
                doc.PathName,
                use_counter,
                used_by_ids,
                main_cat_name,
                subCat.Name,
                subCat.Id.IntegerValue,
                cat_props,
            )
            self.data.append(dic)

    def _build_data_non_main_sub_cats(self, main_sub_cats, main_cat_name, doc):
        """
        Extracts for each category past in, its properties and usage and adds that data to class property .data

        :param  main_sub_cats: List of sub categories to be processed.
        :type  main_sub_cats: [Autodesk.Revit.DB.Category]
        :param main_cat_name: The parent category name.
        :type main_cat_name: str
        :param doc: Current family document
        :type doc: Autodesk.Revit.DB.Document
        """

        # get usage of each main sub category
        for key, sub_category in main_sub_cats.items():
            # set up an empty dic for any sub category not belonging to the main category (no elements in this family can be on those)
            # Exception: the only elements will be Imports and reference planes
            element_dic = {}
            if (
                main_cat_name == "Imports in Families"
                or main_cat_name == "Reference Planes"
            ):
                # get elements using category
                element_dic = rElementByCatUtils.get_elements_by_category(
                    doc, sub_category
                )
            # get category property
            cat_props = rCatPropGet.get_category_properties(sub_category, doc)
            # add element counter for 3D, Cut, Elevation style
            use_counter = 0
            for key in element_dic:
                use_counter = use_counter + len(element_dic[key])
            # add element ids integer value in 3D, Cut, Elevation style
            used_by_ids = []
            for key in element_dic:
                for id in element_dic[key]:
                    used_by_ids.append(id.IntegerValue)

            # build data dictionary
            dic = self._create_data(
                self.root_path,
                self.root_category_path,
                self._strip_file_extension(doc.Title),
                doc.PathName,
                use_counter,
                used_by_ids,
                main_cat_name,
                sub_category.Name,
                sub_category.Id.IntegerValue,
                cat_props,
            )
            self.data.append(dic)

    def process(self, doc):
        """
        Collects all category data from the document and stores it in the class property .data

        :param doc: Current family document
        :type doc: Autodesk.Revit.DB.Document
        """

        # get the family category name:
        fam_cat_name = list(rCat.get_family_category(doc))[0]
        # get all sub categories of the family category
        main_sub_cats = rCat.get_main_sub_categories(doc)
        # get all sub categories of non family category with a positive Id (indicates a custom category)
        # this include imported element categories
        other_custom_sub_cats = rCat.get_other_custom_sub_categories(doc)
        # get usage of each main sub category
        self._build_data(main_sub_cats, fam_cat_name, doc)
        # add any other sub category if exist
        if len(other_custom_sub_cats) > 0:
            for category_key, otherSubMainCat in other_custom_sub_cats.items():
                if len(otherSubMainCat) > 0:
                    # get usage of each other sub category
                    self._build_data_non_main_sub_cats(
                        otherSubMainCat, category_key, doc
                    )

    def get_data(self):
        return self.data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to purging Revit building pad types. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Common.purge_utils import get_used_unused_type_ids
from duHast.Revit.BuildingPads.Utility import (
    RevitBuildingPadTypeSorting as rBuildingPadSort,
)
from duHast.Revit.BuildingPads.building_pads import (
    get_all_building_pad_type_ids_in_model_by_category,
    get_all_building_pad_type_ids_in_model_by_class,
)

#: Built in family name for pad
BASIC_BUILDING_PAD_FAMILY_NAME = "Pad"

#: List of all Built in pad family names
BUILTIN_BUILDING_PAD_TYPE_FAMILY_NAMES = [BASIC_BUILDING_PAD_FAMILY_NAME]


def get_used_building_pad_type_ids(doc):
    """
    Gets all used building pad type ids.
    Filters by category.
    Used: at least one instance of this type is placed in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing used building pad types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = get_used_unused_type_ids(
        doc, get_all_building_pad_type_ids_in_model_by_category, 1
    )
    return ids


def family_no_types_in_use(fam_type_ids, un_used_type_ids):
    """
    Compares two lists of ids. True if any id is not in unUsedTypeIds.
    TODO: check for more generic list comparison and remove this function.
    :param fam_type_ids: List of family type ids to check.
    :type fam_type_ids: List of Autodesk.Revit.DB.ElementId
    :param un_used_type_ids: Reference list of ids.
    :type un_used_type_ids: List of Autodesk.Revit.DB.ElementId
    :return: True if any id from famTypeIds is not in unUsedTypeIds.
    :rtype: bool
    """

    match = True
    for fam_type_id in fam_type_ids:
        if fam_type_id not in un_used_type_ids:
            match = False
            break
    return match


def get_unused_non_in_place_building_pad_type_ids_to_purge(doc):
    """
    Gets all unused building pad type id's.
    - Basic BuildingPad
    This method can be used to safely delete unused building pad types:
    In the case that no building pad instance using any of the types is placed this will return all but one type id since\
        Revit requires at least one building pad type definition to be in the model.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List of element ids representing not used building pad types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    # get unused type ids
    ids = get_used_unused_type_ids(
        doc, get_all_building_pad_type_ids_in_model_by_class, 0
    )
    # make sure there is at least on BuildingPad type per system family left in model
    building_pad_types = rBuildingPadSort.sort_building_pad_types_by_family_name(doc)
    for key, value in building_pad_types.items():
        if key in BUILTIN_BUILDING_PAD_TYPE_FAMILY_NAMES:
            if family_no_types_in_use(value, ids) == True:
                # remove one type of this system family from unused list
                ids.remove(value[0])
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit stairs. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

from duHast.Revit.Common import common as com
from duHast.Revit.Family import family_utils as rFam
from duHast.Revit.Stairs.Utility import stairs_filter as rStairsFilter

# import Autodesk
import Autodesk.Revit.DB as rdb
import Autodesk.Revit.DB.Architecture as rdbA


def get_all_stair_types_by_category(doc):
    """
    Gets a filtered element collector of all Stair types in the model.

    Return includes:
    - Stair
    - Assembled Stair
    - Precast Stair
    - Cast-In-Place Stair
    - In place families or loaded families

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing stair types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rStairsFilter._get_all_stair_types_by_category(doc)
    return collector


def get_all_stair_types_by_class(doc):
    """
    Gets a filtered element collector of all Stair types in the model.

    Return includes:

    - Assembled Stair
    - Precast Stair
    - Cast-In-Place Stair

    It will not return any in place family or Stair types! These are internally treated as Families or Family Symbols class objects.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing stair types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = rStairsFilter._get_stair_types_by_class(doc)
    return collector


# -------------------------------- none in place Stair types -------------------------------------------------------


def get_all_stair_instances_by_category(doc):
    """
    Gets a filtered element collector of all Stair elements placed in model.

    TODO: Confirm it  ignores in place families?

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing stair instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return (
        rdb.FilteredElementCollector(doc)
        .OfCategory(rdb.BuiltInCategory.OST_Stairs)
        .WhereElementIsNotElementType()
    )


def get_all_stair_instances_by_class(doc):
    """
    Gets a filtered element collection all Stair elements placed in model...

    TODO: Confirm it ignores Stair soffits.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing stair instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    return (
        rdb.FilteredElementCollector(doc)
        .OfClass(rdbA.Stairs)
        .WhereElementIsNotElementType()
    )


def get_all_stair_type_ids_by_category(doc):
    """
    Gets all Stair element type ids available in model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids representing stair types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col_cat = get_all_stair_types_by_category(doc)
    ids = com.get_ids_from_element_collector(col_cat)
    return ids


def get_all_stair_type_ids_by_class(doc):
    """
    Gets all Stair element type ids available in model.

    Ignores in place families of category stair.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids representing stair types.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col_class = get_all_stair_types_by_class(doc)
    ids = com.get_ids_from_element_collector(col_class)
    return ids


# -------------------------------- In place Stair types -------------------------------------------------------


def get_in_place_stair_family_instances(doc):
    """
    Gets all instances in place families of category stair.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing in place stair family instances.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    filter = rdb.ElementCategoryFilter(rdb.BuiltInCategory.OST_Stairs)
    return (
        rdb.FilteredElementCollector(doc)
        .OfClass(rdb.FamilyInstance)
        .WherePasses(filter)
    )


def get_all_in_place_stair_type_ids(doc):
    """
    Gets all type ids off all available in place families of category stair.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: List of element ids
    :rtype: list of Autodesk.Revit.ElementIds
    """

    ids = rFam.get_all_in_place_type_ids_in_model_of_category(
        doc, rdb.BuiltInCategory.OST_Stairs
    )
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit view schedules. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import Autodesk.Revit.DB as rdb

from duHast.Revit.Views.Utility.view_types import _get_view_types


def get_schedule_ids_on_sheets(doc):
    """
    Gets view ids of all schedules with instances placed on a sheet
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: List containing schedule Id's.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col = rdb.FilteredElementCollector(doc).OfClass(rdb.ScheduleSheetInstance)
    for s in col:
        if s.ScheduleId not in ids:
            ids.append(s.ScheduleId)
    return ids


def filter_revision_schedules(view):
    """
    Checks whether a view is a revision schedule.
    (not required...schedules have a property flag!!)

    :param view: The view to check.
    :type view: Autodesk.Revit.DB.View
    :return: True if the view name starts with '<', otherwise False
    :rtype: bool
    """

    if view.Name.startswith("<"):
        return False
    else:
        return True


def get_schedules_not_on_sheets(doc):
    """
    Gets all schedules without an instance placed on a sheet.
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: list of schedules without a sheet schedule instance.
    :rtype: list of Autodesk.Revit.DB.View
    """

    schedules_not_on_sheets = []
    # get schedules on sheets
    ids_on_sheets = get_schedule_ids_on_sheets(doc)
    # get all schedules in model
    schedules_in_model = _get_view_types(doc, rdb.ViewType.Schedule)
    # loop and filter out schedules not on sheets
    for schedule in schedules_in_model:
        if schedule.Id not in ids_on_sheets:
            schedules_not_on_sheets.append(schedule)
    return schedules_not_on_sheets



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A class used to implement deleted element count for line styles when using purge by delete.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

- checks whether the deleted element count is a list of 2 entries
- checks whether the first entry is a line style id
- checks whether the second entry is a graphics style id belonging to the line style id


This class is used to modify the deleted element count for line styles when using purge by delete because this modifier is called
after the elements are deleted from the model but a reference is required of them before they are deleted.
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import System
from System.Collections.Generic import List #required since we are dealing with a c# List element

from duHast.Revit.Purge.Objects.ModifierBase import ModifierBase
from duHast.Revit.LinePattern.line_styles import get_all_graphics_style_ids_by_line_style_id 

# required for isInstance check
from Autodesk.Revit.DB import (
    ElementId,
)
class LineStylePurgeModifier(ModifierBase):
    def __init__(self, doc):
        """
        Class constructor.

        """

        super(ModifierBase, self).__init__()

        # get dictionary of line styles
        self.graphic_style_id_by_line_style_id = get_all_graphics_style_ids_by_line_style_id(doc)
        self.debug_log = []

    def modify_deleted(self, doc, deleted):
        """
        Base implementation override to modify the deleted element count.

        Returns deleted element list unchanged if:
        - deleted is not a c# List[ElementId]
        - deleted is not exactly two ids
        - first deleted id is not in the dictionary of graphic style ids by line style ids
        - second deleted is not equal to the dictionary value of the first deleted id

        Returns:
        - deleted element list with the first id adjusted to 1 if the second id is equal to the dictionary value of the first id

        Args:
            deleted: The deleted element count

        """

        if isinstance(deleted, List[ElementId]) == False:
            self.debug_log.append("deleted is not a List[ElementId]: {}".format(deleted))
            raise TypeError("deleted must be a List[ElementId]:".format(deleted))

        # should be exactly two ids
        if (len(deleted) != 2):
            self.debug_log.append("deleted count is not 2: {}".format(deleted))
            return deleted
        
        # sort ids ascending
        # first id should be the line style and second (higher value) the graphics style
        sorted_ids = sorted(deleted, key=lambda x: x.IntegerValue)
        if(sorted_ids[0].IntegerValue in self.graphic_style_id_by_line_style_id):
            self.debug_log.append("sorted_id: {} is in dictionary: {}".format(sorted_ids[0], self.graphic_style_id_by_line_style_id ))
            if(sorted_ids[1].IntegerValue == self.graphic_style_id_by_line_style_id [sorted_ids[0].IntegerValue]):
                self.debug_log.append("sorted_id: {} is equal to dictionary value: {}".format(sorted_ids[1], self.graphic_style_id_by_line_style_id[sorted_ids[0].IntegerValue] ))
                # adjusted the deleted id value to 1
                return [sorted_ids[0]]
            else:
                self.debug_log.append("sorted_id: {} is not equal to dictionary value: {}".format(sorted_ids[1], self.graphic_style_id_by_line_style_id[sorted_ids[0].IntegerValue] ))
                return deleted
        # leave the deleted id count value as is
        else:  
            self.debug_log.append("sorted_id: {} is not in dictionary: {}".format(sorted_ids[0], self.graphic_style_id_by_line_style_id ))
        return deleted



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
An implementation of the Revit IFailuresProcessor interface
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Helps to pre process warnings in Revit.

"""

# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from Autodesk.Revit.DB import IFailuresPreprocessor

from duHast.Utilities.Objects.result import Result
from duHast.Revit.Common.failure_handling import process_failures
from duHast.Revit.Common.Objects.FailureHandlingConfiguration import (
    FailureHandlingConfig,
)


class FailuresPreprocessor(IFailuresPreprocessor):
    def __init__(
        self,
        failure_processor=process_failures,
        result=Result(),
        fail_config=FailureHandlingConfig(),
    ):
        """
        Constructor for the FailuresPreprocessor class.

        :param failure_processor: The function to be used to process failures.
        :type failure_processor: function
        :param result: The result of the operation.
        :type result: duHast.Utilities.Objects.result.Result
        :param fail_config: The failure handling configuration.
        :type fail_config: duHast.Revit.Common.Objects.FailureHandlingConfiguration.FailureHandlingConfig

        """

        # some checking of the input parameters
        if callable(failure_processor):
            self.failure_processor = failure_processor
        else:
            raise TypeError("failure_processor must be a callable function")

        if isinstance(fail_config, FailureHandlingConfig):
            self.failure_handling_config = fail_config
        else:
            raise TypeError(
                "fail_config must be an instance of duHast.Revit.Common.Objects.FailureHandlingConfiguration.FailureHandlingConfig"
            )

        if isinstance(result, Result):
            self.result = result
        else:
            raise TypeError(
                "result must be an instance of duHast.Utilities.Objects.result.Result"
            )

        return

    def PreprocessFailures(self, failures_accessor):
        # If the failure processing function is the default from the failure_handling module
        # Then we need to pass the failure handling configuration to the function
        if self.failure_processor == process_failures:
            try:
                result = self.failure_processor(
                    failures_accessor,
                    self.failure_handling_config,
                )
            except Exception as e:
                print("exception in failure preprocessor: {}".format(e))
        else:
            result = self.failure_processor(failures_accessor, self.result)
        return result



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A base class used to store view (template) graphic settings .
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


Stores:

- model overrides
- filter overrides


"""


#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import json

from duHast.Utilities.Objects import base

from duHast.Revit.Views.Objects.Data.override_by_category import OverrideByCategory
from duHast.Revit.Views.Objects.Data.override_by_filter import OverrideByFilter


class ViewGraphicsSettings(base.Base):
    def __init__(self, view_name="", view_id=-1, j=None, **kwargs):
        """
        Class constructor.

        """

        super(ViewGraphicsSettings, self).__init__(**kwargs)

        # set default values
        self.view_name = view_name
        self.view_id = view_id
        self.override_by_category = []
        self.override_by_filter = []

        # check if any data was past in with constructor!
        if j != None and len(j) > 0:
            # check type of data that came in:
            if type(j) == str:
                # a string
                j = json.loads(j)
            elif type(j) == dict:
                # no action required
                pass
            else:
                raise ValueError(
                    "Argument supplied must be of type string or type dictionary"
                )

            # load values and throw exception if something is missing!
            try:
                for override in j[OverrideByCategory.data_type]:
                    self.override_by_category.append(OverrideByCategory(j=override))
                for override in j[OverrideByFilter.data_type]:
                    self.override_by_filter.append(OverrideByFilter(j=override))
                self.view_name = j["view_name"]
                self.view_id = j["view_id"]
            except Exception as e:
                raise ValueError(
                    "Node ViewGraphicsSettings failed to initialise with: {}".format(e)
                )

    def get_differing_category_overrides(self, other_view_graphic_settings):
        """
        Returns all category overrides of this category of categories  which exist in both set of settings but have different overrides applied.

        :param other_view_graphic_settings: _description_
        :type other_view_graphic_settings: _type_
        :raises TypeError: _description_
        """

        if not isinstance(other_view_graphic_settings, ViewGraphicsSettings):
            raise TypeError(
                "other_view_graphic_settings must be an instance of ViewGraphicsSettings"
            )

        difference = []

        # check category overrides
        for source_override_category in self.override_by_category:
            for (
                other_override_category
            ) in other_view_graphic_settings.override_by_category:
                if (
                    source_override_category.main_category_name
                    == other_override_category.main_category_name
                    and source_override_category.sub_category_name
                    == other_override_category.sub_category_name
                ):
                    if source_override_category != other_override_category:
                        difference.append(source_override_category)
                    break

        return difference

    def get_differing_filter_overrides(self, other_view_graphic_settings):
        """
        Returns all filter overrides which exist exist in both set of settings but have different overrides applied.

        :param other_view_graphic_settings: _description_
        :type other_view_graphic_settings: _type_
        :raises TypeError: _description_
        """

        if not isinstance(other_view_graphic_settings, ViewGraphicsSettings):
            raise TypeError(
                "other_view_graphic_settings must be an instance of ViewGraphicsSettings"
            )

        difference = []

        # check filter overrides
        for source_override_category in self.override_by_filter:
            for (
                other_override_category
            ) in other_view_graphic_settings.override_by_filter:
                if (
                    source_override_category.filter_name
                    == other_override_category.filter_name
                ):
                    if source_override_category != other_override_category:
                        difference.append(source_override_category)
                    break
        return difference

    def get_all_used_line_patterns(self):
        """
        Get all line patterns used in overrides

        :return: Dictionary of line pattern names to line pattern objects
        :rtype: {str: :class:`.LinePatternSettings`}
        """

        all_line_patterns = {}

        # check category overrides
        for override_category in self.override_by_category:
            line_pattern_override = override_category.get_all_used_line_patterns()
            all_line_patterns.update(line_pattern_override)

        # check filter overrides
        for override_filter in self.override_by_filter:
            line_pattern_override = override_filter.get_all_used_line_patterns()
            all_line_patterns.update(line_pattern_override)
        return all_line_patterns

    def get_all_used_fill_patterns(self):
        """
        Get all fill pattern used in overrides.

        :return: Dictionary of line pattern names to line pattern objects
        :rtype: {str: :class:`.FillPatternSettings`}

        :return: _description_
        :rtype: _type_
        """
        all_fill_patterns = {}

        # check category overrides
        for override_category in self.override_by_category:
            fill_pattern_override = override_category.get_all_used_fill_patterns()
            all_fill_patterns.update(fill_pattern_override)

        # check filter overrides
        for override_filter in self.override_by_filter:
            fill_pattern_override = override_filter.get_all_used_fill_patterns()
            all_fill_patterns.update(fill_pattern_override)
        return all_fill_patterns



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family Base data utility module containing functions to read the data from file.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Reads family base data into two list of named tuples.

root_family:

- name 
- category 
- filePath 
- parent 
- child

nested_family:

- name 
- category 
- filePath 
- rootPath  [str]
- categoryPath [str]

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

from collections import namedtuple

from duHast.Utilities import (
    files_csv as fileCSV,
    files_get as fileGet,
    files_io as fileIO,
    utility as util,
)

# tuples containing base family data read from file
root_family = namedtuple("root_family", "name category filePath parent child report_data")
nested_family = namedtuple(
    "nested_family", "name category filePath rootPath categoryPath hostFamily report_data"
)

# row structure of family base data file
BASE_DATA_LIST_INDEX_FAMILY_NAME = 2
BASE_DATA_LIST_INDEX_FAMILY_FILE_PATH = 3
BASE_DATA_LIST_INDEX_CATEGORY = 4
BASE_DATA_LIST_INDEX_ROOT_PATH = 0
BASE_DATA_LIST_INDEX_ROOT_CATEGORY_PATH = 1

# file name identifiers for family base data
FAMILY_BASE_DATA_FILE_NAME_PREFIX = "FamilyBase"
FAMILY_BASE_DATA_FILE_EXTENSION = ".csv"

# exceptions
EXCEPTION_NO_FAMILY_BASE_DATA_FILES = "Families base data list files do not exist."
EXCEPTION_EMPTY_FAMILY_BASE_DATA_FILES = "Empty Families base data list file!"


def _get_base_data_file_name(directory_path):
    """
    Gets the first family base data file in provided directory or any of it's sub directories.

    :param directory_path: Fully qualified directory path.
    :type directory_path: str
    :raises Exception: EXCEPTION_NO_FAMILY_BASE_DATA_FILES

    :return: Fully qualified file path to family base data file.
    :rtype: str
    """

    # get all base data files in folder
    files = fileGet.get_files_from_directory_walker_with_filters(
        directory_path,
        FAMILY_BASE_DATA_FILE_NAME_PREFIX,
        "",
        FAMILY_BASE_DATA_FILE_EXTENSION,
    )
    if len(files) > 0:
        return files[0]
    else:
        raise Exception(EXCEPTION_NO_FAMILY_BASE_DATA_FILES)


def read_overall_family_data_list(file_path):
    """
    Reads list of families from family base data report file into named tuples.

    :param file_path: Fully qualified file path to family base data report file.
    :type file_path: str
    :raises Exception: "Families base data list files does not exist."
    :raises Exception: "Empty Families base data list file!"
    :return: Two lists: first list of named tuples contain family root data, second list contains family nested data.
    :rtype: [root_family], [nested_family]
    """

    rows = []
    if fileIO.file_exist(file_path):
        rows = fileCSV.read_csv_file(file_path)
    else:
        raise Exception(EXCEPTION_NO_FAMILY_BASE_DATA_FILES)
    if len(rows) > 0:
        pass
    else:
        raise Exception(EXCEPTION_EMPTY_FAMILY_BASE_DATA_FILES)

    return_value_root_family = []
    return_value_nested_family = []
    for i in range(1, len(rows)):
        # check if root family
        if "::" not in rows[i][BASE_DATA_LIST_INDEX_ROOT_PATH]:
            data = root_family(
                rows[i][BASE_DATA_LIST_INDEX_FAMILY_NAME],
                rows[i][BASE_DATA_LIST_INDEX_CATEGORY],
                rows[i][BASE_DATA_LIST_INDEX_FAMILY_FILE_PATH],
                [],  # set up an empty list for parent families
                [],  # set up an empty list for child families
                rows[i],  # set up report data
            )
            return_value_root_family.append(data)
        else:
            # found a child family
            data = nested_family(
                rows[i][BASE_DATA_LIST_INDEX_FAMILY_NAME],
                rows[i][BASE_DATA_LIST_INDEX_CATEGORY],
                rows[i][BASE_DATA_LIST_INDEX_FAMILY_FILE_PATH],
                rows[i][BASE_DATA_LIST_INDEX_ROOT_PATH].split(
                    " :: "
                ),  # split root path into list for ease of searching
                rows[i][BASE_DATA_LIST_INDEX_ROOT_CATEGORY_PATH].split(
                    " :: "
                ),  # split category path into list for ease of searching
                [],
                rows[i],  # set up report data
            )
            return_value_nested_family.append(data)
    return return_value_root_family, return_value_nested_family


def read_overall_family_data_list_from_directory(directory_path):
    """
    Reads the first family base data file it finds in a folder.
    Note: This method calls ReadOverallFamilyDataList(filePath) which will raise exceptions if files are empty or dont exist in specified folder.

    :param directory_path: A fully qualified directory path containing family base data file(s)
    :type directory_path: _str

    :return: Two lists: first list of named tuples contain family root data, second list contains family nested data.
    :rtype: [root_family], [nested_family]
    """

    file_name = _get_base_data_file_name(directory_path)
    return read_overall_family_data_list(file_name)


def read_overall_family_data_list_into_nested(file_path):
    """
    Reads list of families from family base data report file into named tuples.

    :param file_path: Fully qualified file path to family base data report file.
    :type file_path: str
    :raises Exception: "Families base data list files does not exist."
    :raises Exception: "Empty Families base data list file!"
    :return: A list contains family nested data.
    :rtype: [nested_family]
    """

    rows = []
    if fileIO.file_exist(file_path):
        rows = fileCSV.read_csv_file(file_path)
    else:
        raise Exception(EXCEPTION_NO_FAMILY_BASE_DATA_FILES)
    if len(rows) > 0:
        pass
    else:
        raise Exception(EXCEPTION_EMPTY_FAMILY_BASE_DATA_FILES)

    return_value_nested_family = []
    for i in range(1, len(rows)):
        # found a child family
        data = nested_family(
            rows[i][BASE_DATA_LIST_INDEX_FAMILY_NAME],
            rows[i][BASE_DATA_LIST_INDEX_CATEGORY],
            rows[i][BASE_DATA_LIST_INDEX_FAMILY_FILE_PATH],
            rows[i][BASE_DATA_LIST_INDEX_ROOT_PATH].split(
                " :: "
            ),  # split root path into list for ease of searching
            rows[i][BASE_DATA_LIST_INDEX_ROOT_CATEGORY_PATH].split(
                " :: "
            ),  # split category path into list for ease of searching
            [],
        )
        return_value_nested_family.append(data)
    return return_value_nested_family


def read_overall_family_data_list_into_nested_from_directory(directory_path):
    """
    Reads the first family base data file it finds in a folder.
    Note: This method calls ReadOverallFamilyIntoNestedDataList(filePath) which will raise exceptions if files are empty or dont exist in specified folder.

    :param directory_path: A fully qualified directory path containing family base data file(s)
    :type directory_path: _str

    :return: A list contains family nested data.
    :rtype: [nested_family]
    """

    file_name = _get_base_data_file_name(directory_path)
    return read_overall_family_data_list_into_nested(file_name)


# -------------------------------- simplify data set ----------------------------------------------------------------


def _check_data_blocks_for_overlap(block_one, block_two):
    """
    Checks whether the root path of families in the first block overlaps with the root path of any family in the second block.
    Overlap is checked from the start of the root path. Any families from block one which are not overlapping any family in\
        block two are returned.

    :param block_one: List of family tuples of type nested_family
    :type block_one: [nested_family]
    :param block_two: List of family tuples of type nested_family
    :type block_two: [nested_family]
    
    :return: List of family tuples of type nested_family
    :rtype: [nested_family]
    """

    unique_tree_nodes = []
    for fam in block_one:
        match = False
        for fam_up in block_two:
            if " :: ".join(fam_up.rootPath).startswith(" :: ".join(fam.rootPath)):
                match = True
                break
        if match == False:
            unique_tree_nodes.append(fam)
    return unique_tree_nodes


def _cull_data_block(family_base_nested_data_block):
    """
    Sorts family data blocks into a dictionary where key, from 1 onwards, is the level of nesting indicated by number of '::' in root path string.

    After sorting it compares adjacent blocks in the dictionary (key and key + 1) for overlaps in the root path string. Only unique families will be returned.

    :param family_base_nested_data_block: A list containing all nested families belonging to a single root host family.
    :type family_base_nested_data_block: [nested_family]

    :return: A list of unique families in terms of root path.
    :rtype: [nested_family]
    """

    culled_family_base_nested_data_blocks = []

    if isinstance(family_base_nested_data_block, list) == False:
        raise Exception(
            "Family base nested data block is not a list. It is of type: {}".format(
                type(family_base_nested_data_block)
            )
        )

    culled_family_base_nested_data_blocks = []
    data_blocks_by_length = {}
    
    # build dic by root path length
    # start at 1 because for nesting level ( 1 based rather then 0 based )
    for family in family_base_nested_data_block:

        # root path is a list of family names representing the nesting tree
        # subtract 1 to get the nesting level minus the root family itself
        split_root_path_length = len(family.rootPath) - 1

        # add family to dictionary
        if split_root_path_length in data_blocks_by_length:
            data_blocks_by_length[split_root_path_length].append(family)
        else:
            data_blocks_by_length[split_root_path_length] = [family]
    
    # loop over dictionary and check block entries against next entry up blocks
    # I need to extend range by 1 since the end value in the range is always exclusive in python for i in range loop
    # i.e. for i in range (1, 3) will only loop over 1 and 2...
    
    for i in range(1, len(data_blocks_by_length) + 1):

        # last block get automatically added
        # do not add + 1 here since the loop will stop at the last entry
        if i == len(data_blocks_by_length):
            culled_family_base_nested_data_blocks = (
                culled_family_base_nested_data_blocks + data_blocks_by_length[i]
            )
        else:
            # check for matches in next one up
            unique_nodes = _check_data_blocks_for_overlap(
                data_blocks_by_length[i], data_blocks_by_length[i + 1]
            )
            # only add non overlapping blocks
            culled_family_base_nested_data_blocks = (
                culled_family_base_nested_data_blocks + unique_nodes
            )
    
    return culled_family_base_nested_data_blocks


def cull_nested_base_data_blocks(root_families, nested_families):
    """
    Reduce base data families for parent / child finding purposes. Keep the nodes with the root path longes branch only.

    Sample ( within a single data block)

    famA :: famB :: famC
    FamA :: famB

    The second of the above examples can be culled since the path is contained within the first example.

    Sample ( across multiple data blocks)

    famA :: famB :: famC
    famB :: famC


    The second of the above examples can be culled since the first family (famB) in that sample is already in the first example.(nested Family)

    :param overall_family_base_nested_data: _description_
    :type overall_family_base_nested_data: _type_
    """

    # get root families not found in nested data since they contain the longest unique nesting path
    non_nested_root_families = find_root_families_not_in_nested_data(
        root_families, nested_families
    )

    # get nested family data blocks for each root family
    nested_data_blocks_per_root_family = get_nested_family_data_for_root_family(
        non_nested_root_families, nested_families
    )

    # storage for culled data
    retained_family_base_nested_data = []

    # cull data per block
    for (
        root_family_identifier,
        family_block,
    ) in nested_data_blocks_per_root_family.items():
        # if the data block only contains one item skip culling
        if len(family_block) == 1:
            # no need to do any culling
            retained_family_base_nested_data.append(family_block[0])
            continue
        else:
            # attempt to cull data block
            d = _cull_data_block(family_block)
            retained_family_base_nested_data = retained_family_base_nested_data + d

    return retained_family_base_nested_data


def get_nested_family_data_for_root_family(root_families, nested_families):
    """
    Returns a list of nested families for each root family.

    :param root_families: A list of tuples containing root family data.
    :type root_families: [root_family]
    :param nested_families: A list of tuples containing nested family data.
    :type nested_families: [nested_family]

    :return: A dictionary where:

        - key is the root family name and category concatenated and
        - value is a list of tuples containing nested family data.

    :rtype: {str: [nested_family]}
    """

    nested_families_for_root_families = {}
    for root_family in root_families:
        nested_families_for_root_families[root_family.name + root_family.category] = []
        for nested_family in nested_families:
            if (
                nested_family.rootPath[0] == root_family.name
                and nested_family.categoryPath[0] == root_family.category
            ):
                nested_families_for_root_families[
                    root_family.name + root_family.category
                ].append(nested_family)
    return nested_families_for_root_families


def find_root_families_not_in_nested_data(root_families, nested_families):
    """
    Returns a list of root families not found in the nested family data set.

    In order to reduce the number of families to be processed, this function can be used to find the root families which are not nested into any other family and therefore do contain the
    longest nesting path.

    This is based on two assumptions:

    - The provided data set contains all families in the project (no missing families are reported)
    - As soon as a family is nested into another family, its nesting path will appear in the nested family data set and the root family data set.

    :param root_families: A list of tuples containing root family data.
    :type root_families: [root_family]
    :param nested_families: A list of tuples containing nested family data.
    :type nested_families: [nested_family]

    :return: A list of tuples containing root family data.
    :rtype: [root_family]

    """

    # storage for root families not found in nested data
    root_families_not_in_nested_data = []

    # check each root family
    for root_family in root_families:
        found = False
        for nested_family in nested_families:
            # check if the root family name is in the nested family root path on first position
            # only interested in matches which are not the first entry in the list since that indicates nesting
            index_match_root = util.index_of(nested_family.rootPath, root_family.name)
            if index_match_root > 0:
                # check if the categories are also a match!
                if nested_family.categoryPath[index_match_root] == root_family.category:
                    # found a match
                    found = True
                    break
        if found == False:
            root_families_not_in_nested_data.append(root_family)
    return root_families_not_in_nested_data


# --------------------------------------------  find families in nesting tree data ------------------------------------


def find_direct_host_families(nested_fam, overall_family_base_nested_data):
    """
    Finds the direct hosts of the past in family in the base nested family data set.

    :param nested_fam: A tuple containing nested family data.
    :type nested_fam: nested_family
    :param overall_family_base_nested_data: List of tuples containing nested family data.
    :type overall_family_base_nested_data: [nested_family]

    :return: A dictionary where:

        - key is the family name and category concatenated and
        - value is a tuple in format 0: family name, 1: family category

    :rtype: {str: (str,str)}
    """

    host_families = {}
    # check each base family data whether it contains the missing family in its nesting tree
    for base_nested_fam in overall_family_base_nested_data:
        if nested_fam.name in base_nested_fam.rootPath:
            index_match = util.index_of(base_nested_fam.rootPath, nested_fam.name)
            # make sure we have a match and it is not the first entry in list (does not have a parent...)
            if index_match > 0:
                # confirm category is the same
                if base_nested_fam.categoryPath[index_match] == nested_fam.category:
                    # got a direct parent! (index - 1)
                    key_new = (
                        base_nested_fam.rootPath[index_match - 1]
                        + base_nested_fam.categoryPath[index_match - 1]
                    )
                    if key_new not in host_families:
                        host_families[key_new] = (
                            base_nested_fam.rootPath[index_match - 1],
                            base_nested_fam.categoryPath[index_match - 1],
                        )
    return host_families


def find_all_direct_host_families(families, overall_family_base_nested_data):
    """
    Returns a dictionary of all direct host families of families past in.

    :param families: A list of tuples containing nested family data.
    :type families: [nested_family]
    :param overall_family_base_nested_data: List of tuples containing nested family data.
    :type overall_family_base_nested_data: [nested_family]

    :return: A dictionary.
    :rtype: {str: (str,str)}
    """

    host_families = {}
    for fam in families:
        hosts = find_direct_host_families(fam, overall_family_base_nested_data)
        # update dictionary with new hosts only
        for h in hosts:
            if h not in host_families:
                host_families[h] = hosts[h]
    return host_families


def find_root_families_from_hosts(host_families, overall_family_base_root_data):
    """
    Returns a list of tuples of type root_family matching the past in host families.

    :param host_families: A dictionary where:

        - key is the family name and category concatenated and
        - value is a tuple in format 0: family name, 1: family category

    :type host_families: {str: (str,str)}
    :param overall_family_base_root_data: List of tuples containing root family data.
    :type overall_family_base_root_data: [root_family]

    :return: List of root family tuples.
    :rtype: [root_family]
    """

    base_host_families = []
    for nested_id, nestedFam in host_families.items():
        for base_root_fam in overall_family_base_root_data:
            if (
                nestedFam[0] == base_root_fam.name
                and nestedFam[1] == base_root_fam.category
            ):
                base_host_families.append(base_root_fam)
    return base_host_families



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions to sort task evenly into workload buckets.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A bucket contains a number of files which will eventually be written into task files to be processed by batch processor.
The below function attempt to fill these buckets evenly measured on file size.
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed. 
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits; 
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

from duHast.UI.Objects import workload_bucket as wb


def distribute_workload(number_of_buckets, items, getWorkloadSize):
    """
    Distributes a given number of items evenly by workload size into workload buckets.

    :param numberOfBuckets: The number of buckets items are to be distributed to
    :type numberOfBuckets: int
    :param items: A list of items.
    :type items: [foo]
    :param getWorkloadSize: A function returning the workload size from an item.
    :type getWorkloadSize: func(foo) -> int

    :return: A list of workload bucket objects containing items.
    :rtype: list[ :class:`.WorkloadBucket`]
    """

    workload_buckets = []
    try:
        # ini bucket list
        for x in range(number_of_buckets):
            workload_buckets.append(wb.WorkloadBucket())
        itemToWorkLoadValues = []

        # build key value list of items and their workload value
        for item in items:
            itemToWorkLoadValues.append([item, getWorkloadSize(item)])

        # sort list by workload size in descending order (biggest item first)
        itemToWorkLoadValues = sort(itemToWorkLoadValues)

        # load up with buckets
        for bucket in itemToWorkLoadValues:
            # find bucket with smallest work load value
            lowBucket = min(
                workload_buckets, key=lambda work_bucket: work_bucket.workload_value
            )
            # add new item to bucket list
            lowBucket.add_item(bucket[0])
            # increase work bucket size by size of item added
            lowBucket.set_workload_value(lowBucket.workload_value + bucket[1])

    except Exception as e:
        print(e)
        pass
    # send loaded buckets back
    return workload_buckets


def sort(sub_li):
    """
    Python code to sort the tuples using second element of sublist. Inplace way to sort using sort().

    Note: not sure a one liner warrants a method...?

    :param sub_li: _description_
    :type sub_li: _type_
    :return: _description_
    :rtype: _type_
    """
    # reverse = None (Sorts in Ascending order)
    # key is set to sort using second element of
    # sublist lambda has been used
    sub_li.sort(key=lambda x: x[1], reverse=True)
    return sub_li
